from typing import Union


BROKER_REDIS = "redis"

class MessageBroker:
    """Abstract class for message brokers."""
    def __init__(self, user: str, password: str, host: str, port: Union[int, str], db: Union[int, str]):
        self.user = user
        self.password = password
        self.host = host
        self.port = port
        self.db = db
        self.broker_name = None

    def connect(self):
        """Connect to the broker, and return the connection object."""
        pass

    def get_broker_name(self):
        """Return the name of the broker."""
        pass
    


import datetime
import json
import redis

from negotium.brokers.main import MessageBroker, BROKER_REDIS
from negotium.mq.trackers import _MessageTracker
from negotium.mq.trackers import _COMMAND_REDIS_ZREM, _COMMAND_REDIS_LREM, _COMMAND_REDIS_BLPOP
from negotium.conf import (
    _MESSAGE_MAIN, _MESSAGE_SCHEDULER, _MESSAGE_SCHEDULER_SORTED_SET, _MESSAGE_PERIODIC_TASKS
)
from negotium.schedules.crontab import Crontab
from negotium.utils.logger import log

class _Publisher:
    def __init__(self, broker: MessageBroker, app_name: str, logfile: str=None):
        self.broker = broker
        self.connection = None
        self._tracker = None
        self.app_name = app_name
        self.logfile = logfile

    def _create_connection(self):
        """Create a connection to the message broker
        """
        self.connection = self.broker.connect()
        self._tracker = _MessageTracker(self.broker, self.app_name)

    def _close_connection(self):
        """Close the connection
        """
        self.connection.close()

    def _publish(self, data: dict, eta: datetime.datetime=None, cron: Crontab=None) -> str:
        """Publish a message to the queue and return the message id
        """
        log(self.logfile, data.get('app_name'), f"Received task: {data.get('function_name')}")
        if self.broker.get_broker_name() == BROKER_REDIS:
            if eta:
                data = {
                    '_task': data,
                    '_eta': eta.strftime('%Y-%m-%d %H:%M:%S.%f')
                }
                self.connection.rpush(_MESSAGE_SCHEDULER + "__" + self.app_name, json.dumps(data))
                message_id = self._tracker._track(
                    name=_MESSAGE_SCHEDULER + "__" + self.app_name, 
                    identifier=json.dumps(data), 
                    command=_COMMAND_REDIS_LREM
                )

                timestamp = datetime.datetime.strptime(eta.strftime('%Y-%m-%d %H:%M:%S.%f'), '%Y-%m-%d %H:%M:%S.%f').timestamp()
                self.connection.zadd(_MESSAGE_SCHEDULER_SORTED_SET + "__" + self.app_name, {json.dumps(data): timestamp})
                return self._tracker._track(
                    name=_MESSAGE_SCHEDULER_SORTED_SET + "__" + self.app_name,
                    identifier=json.dumps(data),
                    command=_COMMAND_REDIS_ZREM,
                    uuid_=message_id
                )
            elif cron:
                data = {
                    '_task': data,
                    '_cron': cron.__str__()
                }
                self.connection.rpush(_MESSAGE_PERIODIC_TASKS + "__" + self.app_name, json.dumps(data))
                return self._tracker._track(
                    name=_MESSAGE_PERIODIC_TASKS + "__" + self.app_name,
                    identifier=json.dumps(data), 
                    command=_COMMAND_REDIS_LREM
                )
            else:
                self.connection.rpush(_MESSAGE_MAIN + "__" + self.app_name, json.dumps(data))
                return self._tracker._track(
                    name=_MESSAGE_MAIN + "__" + self.app_name,
                    command=_COMMAND_REDIS_BLPOP
                )
        else:
            raise NotImplementedError("Broker not implemented")



class Crontab:
    """Crontab class.

    This class is used to create a crontab object that can be used to
    schedule tasks to be executed in a periodic manner.

    Example:
        from negotium.schedules import Crontab

        # create a crontab for every January 1st at midnight
        crontab = Crontab(month=1, day=1, hour=0, minute=0)

        # create a crontab for every 1st day of the month at midnight
        crontab = Crontab(day=1, hour=0, minute=0)

        # create a crontab for every Monday at midnight
        crontab = Crontab(weekday=0, hour=0, minute=0)

        # create a crontab for every midnight
        crontab = Crontab(hour=0, minute=0)

        # create a crontab for every 5 minutes
        crontab = Crontab(minute=5)

        # create a crontab for every 5 seconds
        crontab = Crontab(second=5)

        # using a crontab expression
        crontab = Crontab(expression='1 * * * *') # <-- every hour at minute 1
    """ 
    def __init__(self, month: int=None, day: int=None, weekday: int=None, hour: int=None, minute: int=None, expression: str=None):
        if not month and not day and not weekday and not hour and not minute and not expression:
            raise ValueError('provide at least one argument')
        if expression:
            self.expression = expression
        else:
            self._parse_init(month, day, weekday, hour, minute)

    def __repr__(self):
        return self.expression

    def __str__(self):
        return self.expression

    def _parse_init(self, month: int=None, day: int=None, weekday: int=None, hour: int=None, minute: int=None):
        """Convert the arguments to a crontab expression
        """
        self.expression = "%s %s %s %s %s" % (
            minute if minute is not None else '*',
            hour if hour is not None else '*',
            day if day is not None else '*',
            month if month is not None else '*',
            weekday if weekday is not None else '*'
        )



import datetime
import inspect
from functools import wraps

from .conf import DEFAULT_HOST, DEFAULT_PORT
from .task import _delay, _apply_async, _apply_periodic_async
from negotium.brokers.main import MessageBroker
from negotium.mq.consumer import _Consumer
from negotium.mq.publisher import _Publisher
from negotium.utils.logger import log


class Negotium:
    """This class is the entry point to the negotium application.

    Negotium currently supports Redis as the message broker. This class
    provides a decorator to mark functions as tasks and a method to consume
    tasks from the message broker.

    Example:
        from negotium import Negotium

        negotium = Negotium(app_name="test_app", broker_url="redis://localhost:6379/0")
        negotium.start()

        @negotium.task
        def add(x, y):
            return x + y

    Note: This class should be instantiated at the entry point of your application.
    """
    def __init__(self, app_name: str, broker: MessageBroker, logfile: str=None):
        self.app_name = app_name
        if not self.app_name:
            raise ValueError("app_name must be set")

        if not broker.get_broker_name():
            raise ValueError("invalid broker")

        self.consumer = _Consumer(broker=broker, app_name=app_name, logfile=logfile)
        self.publisher = _Publisher(broker=broker, app_name=app_name, logfile=logfile)
        self.logfile = logfile

    def start(self, *args, **kwargs):
        """Use this method to consume tasks from the message broker

        Example:
            negotium.start()

        Note: This method should be called at the entry point of your application.
        """
        self.consumer.run()

    def close(self, *args, **kwargs):
        """Use this method to close the connection to the message broker

        Example:
            negotium.close()

        Note: This method should be called at the exit point of your application.
        """
        self.consumer.close()

    def task(self, func):
        """Decorator for task functions

        Example:
            @negotium.task
            def add(x, y):
                return x + y
        """
        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        d = {
            'app_name': self.app_name,
            'package_dir': '/'.join(inspect.getfile(func).split('/')[:-1]),
            'package_name': inspect.getfile(func).split('/')[-2],
            'module_name': inspect.getmodulename(inspect.getfile(func)),
            'function_name': func.__name__,
            'timestamp': datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        }
        wrapper.delay = lambda *args, **kwargs: _delay(self.publisher, {**d, 'args': args, 'kwargs': kwargs})
        wrapper.apply_async = lambda eta, args: _apply_async(self.publisher, {**d,'args': args}, eta)

        def apply_periodic_async(cron, args):
            uuid_ = _apply_periodic_async(self.publisher, {**d, 'args': args}, cron)
            # restart the consumer to pick up the new periodic task
            self.consumer._consume_periodic_tasks()
            return uuid_
        wrapper.apply_periodic_async = apply_periodic_async
        return wrapper

    def cancel(self, uuid_: str):
        """Cancel a scheduled task

        Example:
            negotium.cancel(uuid_)
        """
        self.consumer._delete_message(uuid_)



import redis

from typing import Union
from .main import MessageBroker, BROKER_REDIS


class Redis(MessageBroker):
    """Redis message broker."""
    def __init__(self, host: str, port: Union[int, str], db: int=0, user: str="", password: str=""):
        super().__init__(user, password, host, port, db)
        self.broker_name = BROKER_REDIS

    def connect(self) -> redis:
        """Connect to Redis."""
        return redis.Redis(
            host=self.host, port=self.port, db=self.db, username=self.user, password=self.password)

    def get_broker_name(self) -> str:
        return self.broker_name



import json
import redis
import uuid

from negotium import conf
from negotium.brokers.main import MessageBroker, BROKER_REDIS

_COMMAND_REDIS_ZREM = 0
_COMMAND_REDIS_LREM = 1
_COMMAND_REDIS_BLPOP = 2

class _MessageTracker:
    """
    A class to track every incoming and outgoing messages
    """
    def __init__(self, broker: MessageBroker, app_name: str):
        self.broker = broker
        self.connection = broker.connect()
        self.app_name = app_name

    def _track(self, command: int, name: str, identifier: str='', uuid_: str='') -> str:
        """Track a message and return the uuid

        Args:
            name (str): name of the queue
            identifier (str): identifier of the message (e.g. a json string value)
            command (int): command to execute
            uuid_ (str): uuid to use (optional: if not provided, a new uuid will be generated)
        """
        uuid_ = uuid_ or str(uuid.uuid4())
        if self.broker.get_broker_name() == BROKER_REDIS:
            self.connection.lpush(conf._MESSAGE_TRACKER + "__" + self.app_name + "__" + uuid_, json.dumps({
                '_name': name,
                '_identifier': identifier,
                '_command': command
            }))
            return uuid_
        else:
            raise NotImplementedError("Broker not implemented")

    def _delete(self, uuid_: str):
        """Delete a message from the tracker
        """
        if self.broker.get_broker_name() == BROKER_REDIS:
            messages = self.connection.lrange(conf._MESSAGE_TRACKER + "__" + self.app_name + "__" + uuid_, 0, -1)
            for message in messages:
                data = json.loads(message)
                if data.get('_command') == _COMMAND_REDIS_ZREM:
                    self.connection.zrem(data.get('_name'), data.get('_identifier'))
                elif data.get('_command') == _COMMAND_REDIS_LREM:
                    self.connection.lrem(data.get('_name'), 0, data.get('_identifier'))
                elif data.get('_command') == _COMMAND_REDIS_BLPOP:
                    self.connection.blpop(data.get('_identifier'))
            self.connection.delete(conf._MESSAGE_TRACKER + "__" + self.app_name + "__" + uuid_)
        else:
            raise NotImplementedError("Broker not implemented")



import croniter
import datetime
import redis
import inspect
import importlib
import json
import os
import signal
import sys
import time
from threading import Thread, Timer

from negotium.brokers.main import MessageBroker, BROKER_REDIS
from negotium.mq.trackers import _MessageTracker
from negotium.conf import (
    _MESSAGE_MAIN, _MESSAGE_SCHEDULER, _MESSAGE_SCHEDULER_SORTED_SET, _MESSAGE_PERIODIC_TASKS
)
from negotium.utils.logger import log


class _Consumer:
    def __init__(self, broker: MessageBroker, app_name: str, logfile: str=None):
        self.broker = broker
        self.connection = broker.connect()
        self._tracker = _MessageTracker(broker, app_name)
        self._is_closed = False
        self.app_name = app_name
        self.logfile = logfile
        self._thread_consume = None
        self._thread_consume_scheduled = None
        self._thread_consume_periodic = None

    def _close_connection(self):
        """Close the connection
        """
        self.connection.close()

    def _delete_message(self, uuid_: str):
        """Delete a message
        """
        self._tracker._delete(uuid_)

    def _consume(self, *args, **kwargs):
        """Consume messages from the queue
        """
        if self.broker.get_broker_name() == BROKER_REDIS:
            while True:
                if self._is_closed:
                    return
                message = self.connection.blpop(_MESSAGE_MAIN + "__" + self.app_name)
                self._callback(message[1])
                time.sleep(1)
        else:
            raise NotImplementedError("Broker not implemented")

    def _consume_scheduled_tasks(self, *args, **kwargs):
        """Load scheduled tasks
        """
        if self.broker.get_broker_name() == BROKER_REDIS:
            while True:
                if self._is_closed:
                    return
                current_time = datetime.datetime.now().timestamp()
                tasks = self.connection.zrangebyscore(_MESSAGE_SCHEDULER_SORTED_SET + "__" + self.app_name, 0, current_time)
                for task in tasks:
                    task = json.loads(task.decode('utf-8'))
                    eta = task.get('_eta')
                    task = task.get('_task')
                    self.connection.zrem(_MESSAGE_SCHEDULER_SORTED_SET + "__" + self.app_name, json.dumps({
                        '_task': task,
                        '_eta': eta
                    }))
                    # execute task
                    self._callback_scheduled(json.dumps(task), eta)
                time.sleep(1)
        else:
            raise NotImplementedError("Broker not implemented")

    def _consume_periodic_tasks(self, *args, **kwargs):
        """Load periodic tasks
        """
        tasks = []
        if self.broker.get_broker_name() == BROKER_REDIS:
            tasks = self.connection.lrange(_MESSAGE_PERIODIC_TASKS + "__" + self.app_name, 0, -1)
        else:
            raise NotImplementedError("Broker not implemented")

        for task in tasks:
            body = json.loads(task.decode('utf-8'))
            cron = body.get('_cron')
            task = body.get('_task')

            # schedule task with the given cron
            croniter_ = croniter.croniter(cron, datetime.datetime.now())
            self._schedule_periodic_task(croniter_, task, cron)

    def _schedule_periodic_task(self, croniter_: croniter.croniter, task: dict, cron: str):
        """Execute a periodic task at the next cron
        """
        eta = croniter_.get_next(datetime.datetime)
        delay = (eta - datetime.datetime.now()).total_seconds()
        Timer(delay, self._callback_periodic, [json.dumps(task), cron]).start()

    def _callback(self, body):
        """Callback function
        """
        self._execute_task(body)

    def _callback_scheduled(self, body, eta):
        """Callback function for scheduled tasks
        """
        self._execute_task(body)
        self.connection.lrem(_MESSAGE_SCHEDULER + "__" + self.app_name, 0, json.dumps({
            '_task': json.loads(body),
            '_eta': eta
        }))

    def _callback_periodic(self, body, cron):
        """Callback function for periodic tasks
        """
        self._execute_task(body)

    def _execute_task(self, body):
        """Execute a task
        """
        # log the message
        # extract dict from bytes
        body = json.loads(body)

        # get function arguments
        app_name = body.get('app_name')
        package_dir = body.get('package_dir')
        package_name = body.get('package_name')
        module_name = body.get('module_name')
        function_name = body.get('function_name')
        args = body.get('args', [])
        kwargs = body.get('kwargs', {})
        is_scheduled = body.get('_is_scheduled')
        log(self.logfile, app_name, 
            f"{'[Scheduled] ' if is_scheduled else ''}Executing (task: {function_name})", level="INFO")
        
        # import
        spec = importlib.util.spec_from_file_location(f"{module_name}.{function_name}", f"{package_dir}/{module_name}.py")
        module = importlib.util.module_from_spec(spec)
        sys.modules[f"{module_name}.{function_name}"] = module
        spec.loader.exec_module(module)
        # execute the function
        function = getattr(module, function_name)
        try:
            res = function(*args, **kwargs)
            log(self.logfile, app_name, 
                f"{'[Scheduled] ' if is_scheduled else ''}Result (task: {function_name}): {res}", level="INFO")
            return res
        except Exception as e:
            log(self.logfile, app_name, 
                f"{'[Scheduled] ' if is_scheduled else ''}Error (task: {function_name}): {e}", level="ERROR")

    def run(self):
        """Run the consumers in a separate threads
        """
        # create threads
        self._thread_consume = Thread(target=self._consume, daemon=True)
        self._thread_consume_scheduled = Thread(target=self._consume_scheduled_tasks, daemon=True)
        self._thread_consume_periodic = Thread(target=self._consume_periodic_tasks, daemon=True)
        # start threads
        self._thread_consume.start()
        self._thread_consume_scheduled.start()
        self._thread_consume_periodic.start()

    def close(self):
        """Close the connection
        """
        self._is_closed = True
        self._close_connection()



import datetime
import json
import warnings

from negotium.mq.publisher import _Publisher
from negotium.mq.consumer import _Consumer
from negotium.schedules import Crontab
from negotium.conf import _is_worker_enabled, _ignore_execution

def _delay(publisher: _Publisher, data: dict) -> str:
    """Execute the task
    """
    if not _is_worker_enabled():
        if _ignore_execution():
            warnings.warn("The worker is not enabled. The task will be ignored")
            return None
        warnings.warn("The worker is not enabled. The task will be executed synchronously")
        return _Consumer(publisher.broker, publisher.app_name)._execute_task(json.dumps(data))

    publisher._create_connection()
    uuid_ = publisher._publish(data)
    publisher._close_connection()
    return uuid_

def _apply_async(publisher: _Publisher, data: dict, eta: datetime.datetime=None) -> str:
    """Schedule a task to be executed in the future
    """
    if not _is_worker_enabled():
        if _ignore_execution():
            warnings.warn("The worker is not enabled. The task will be ignored")
            return None
        warnings.warn("The worker is not enabled. The task will be executed synchronously")
        return _Consumer(publisher.broker, publisher.app_name)._execute_task(json.dumps(data))

    publisher._create_connection()
    uuid_ = publisher._publish(data, eta=eta)
    publisher._close_connection()
    return uuid_

def _apply_periodic_async(publisher: _Publisher, data: dict, cron: Crontab) -> str:
    """Schedule a periodic task to be executed
    """
    if not _is_worker_enabled():
        if _ignore_execution():
            warnings.warn("The worker is not enabled. The task will be ignored")
            return None
        warnings.warn("The worker is not enabled. The task will be executed synchronously")
        return _Consumer(publisher.broker, publisher.app_name)._execute_task(json.dumps(data))

    publisher._create_connection()
    uuid_ = publisher._publish(data, cron=cron)
    publisher._close_connection()
    return uuid_



"""This module contains the settings for the negotium application.
"""
import os

# default settings
DEFAULT_HOST = 'localhost'
DEFAULT_PORT = 6379
DEFAULT_USERNAME = 'guest'
DEFAULT_PASSWORD = 'guest'

# message settings
_MESSAGE_MAIN = 'negotium_queue'
_MESSAGE_SCHEDULER = 'negotium_scheduler_queue'
_MESSAGE_SCHEDULER_SORTED_SET = 'negotium_scheduler_sorted_set'
_MESSAGE_TRACKER = 'negotium_tracker'
_MESSAGE_PERIODIC_TASKS = 'negotium_periodic_tasks'

# logging settings
LOGGING_FORMAT = '[%(asctime)s] [negotium: %(name)s] [%(levelname)s] %(message)s\n'
LOGGING_DATE_FORMAT = '%Y-%m-%d %H:%M:%S'

# worker settings
def disable_worker(ignore_execution: bool=False) -> bool:
    """Disable the worker to execute tasks asynchronously

    Args:
        execute (bool): ignore execution of tasks (default: False)
            This will completely disable the execution of tasks synchronously

    Returns:
        bool: True if successful
    """
    os.environ['NEGOTIUM_WORKER'] = '0'
    if ignore_execution:
        os.environ['NEGOTIUM_WORKER_IGNORE_EXECUTION'] = '1'
    return True

def enable_worker() -> bool:
    """Enable the worker to execute tasks asynchronously
    """
    os.environ['NEGOTIUM_WORKER'] = '1'
    return True

def _is_worker_enabled() -> bool:
    """Check if the worker is enabled
    """
    return os.environ.get('NEGOTIUM_WORKER', '1') == '1'

def _ignore_execution() -> bool:
    """Check if the execution of tasks should be ignored
    """
    return os.environ.get('NEGOTIUM_WORKER_IGNORE_EXECUTION', '0') == '1'



from datetime import datetime

from negotium.conf import LOGGING_FORMAT, LOGGING_DATE_FORMAT

def log(logfile: str, app_name: str, message: str, level: str="INFO"):
    """Log a message to a file if a logfile is provided.
    If a logfile is not provided, the message is logged to stdout.

    Args:
        app_name (str): The name of the application
        message (str): The message to log
        level (str): The logging level
    """
    if logfile:
        # create a log file
        log_file = open(logfile, 'a')
        # write the message to the log file
        log_file.write(LOGGING_FORMAT % {
            'asctime': datetime.now().strftime(LOGGING_DATE_FORMAT),
            'name': app_name,
            'levelname': level,
            'message': message
        })
        # close the log file
        log_file.close()
    else:
        # log to stdout
        print(LOGGING_FORMAT.replace('\n', '') % {
            'asctime': datetime.now().strftime(LOGGING_DATE_FORMAT),
            'name': app_name,
            'levelname': level,
            'message': message
        })
